import dataclasses
import inspect
import logging
from typing import Any, Dict, List, Optional

import torch
from torch import nn

from .graph import TorchGraph, Node
from .graph_utils import trace_execution
from .utils import sanitize_name


@dataclasses.dataclass
class GraphExecutorState:
    """State for the GraphExecutor that keeps track of both the state of hidden units
    and caches the output of previous modules, for use in (future) recurrent
    computations."""

    state: Dict[str, Any] = dataclasses.field(default_factory=dict)
    cache: Dict[str, Any] = dataclasses.field(default_factory=dict)


class GraphExecutor(nn.Module):
    """Executes the NIR graph in PyTorch.

    By default the graph executor is stateful, since there may be recurrence or
    stateful modules in the graph. Specifically, that means accepting and returning a
    state object (`GraphExecutorState`). If that is not desired,
    set `return_state=False` in the constructor.

    Arguments:
        graph (Graph): The graph to execute
        return_state (bool, optional): Whether to return the state object.
            Defaults to True.

    Raises:
        ValueError: If there are no edges in the graph
    """

    def __init__(self, graph: TorchGraph, return_state: bool = True) -> None:
        super().__init__()
        self.graph = graph
        self.stateful_modules = set()
        self.return_state = return_state
        self.instantiate_modules()
        self.execution_order = self.get_execution_order()
        if len(self.execution_order) == 0:
            raise ValueError("Graph is empty")

    def _is_module_stateful(self, module: torch.nn.Module) -> bool:
        signature = inspect.signature(module.forward)
        arguments = len(signature.parameters)
        # HACK for snntorch modules
        if "snntorch" in str(module.__class__):
            if module.__class__.__name__ in [
                "Synaptic",
                "RSynaptic",
                "Leaky",
                "RLeaky",
            ]:
                return not module.init_hidden
        # Note that GraphExecutor is *also* a stateful module
        return "state" in signature.parameters and arguments > 1

    def get_execution_order(self) -> List[Node]:
        """Evaluate the execution order and instantiate that as a list."""
        inputs = set()
        name_dict = {name: node for node, name in self.graph.module_names.items()}
        for input_node in self.graph.inputs:
            if input_node.name in name_dict:
                inputs.add(input_node)
            else:
                logging.warning(
                    f"Input node {input_node.name} not found in module names. Skipping."
                )
        assert len(inputs) > 0, "No input nodes found, we require at least one."
        if len(inputs) > 1:
            logging.warning(
                "Multiple input nodes found. Using the first node as the input."
            )
        return trace_execution(next(iter(inputs)), lambda n: n.outgoing_nodes.keys())

    def instantiate_modules(self):
        for mod, name in self.graph.module_names.items():
            if mod is not None:
                self.add_module(sanitize_name(name), mod)
                if self._is_module_stateful(mod):
                    self.stateful_modules.add(sanitize_name(name))

    def get_input_nodes(self) -> List[Node]:
        # NOTE: This is a hack. Should use the input nodes from NIR graph
        return self.graph.get_root()

    def _apply_module(
        self,
        node: Node,
        input_nodes: List[Node],
        new_state: GraphExecutorState,
        old_state: GraphExecutorState,
        data: Optional[torch.Tensor] = None,
    ):
        """Applies a module and keeps track of its state.

        TODO: Use pytree to recursively construct the state
        """
        inputs = []
        # Append state if needed
        if node.name in self.stateful_modules and node.name in old_state.state:
            inputs.extend(old_state.state[node.name])

        # Sum recurrence if needed
        summed_inputs = [] if data is None else [data]
        for input_node in input_nodes:
            if (
                input_node.name not in new_state.cache
                and input_node.name in old_state.cache
            ):
                summed_inputs.append(old_state.cache[input_node.name])
            elif input_node.name in new_state.cache:
                summed_inputs.append(new_state.cache[input_node.name])

        if len(summed_inputs) == 0:
            raise ValueError("No inputs found for node {}".format(node.name))
        elif len(summed_inputs) == 1:  # Prepend the input
            inputs.insert(0, summed_inputs[0])
        elif len(summed_inputs) > 1:  # Prepend the sum of the inputs
            inputs.insert(0, torch.stack(summed_inputs).sum(0))

        out = node.elem(*inputs)
        # If the module is stateful, we know the output is (at least) a tuple
        # HACK to make it work for snnTorch
        is_rsynaptic = "snntorch._neurons.rsynaptic.RSynaptic" in str(
            node.elem.__class__
        )
        if is_rsynaptic and not node.elem.init_hidden:
            assert "lif" in node.name, "this shouldnt happen.."
            new_state.state[node.name] = out  # snnTorch requires output inside state
            out = out[0]
        elif node.name in self.stateful_modules:
            new_state.state[node.name] = out[1:]  # Store the new state
            out = out[0]
        return out, new_state

    def forward(self, data: torch.Tensor, state: Optional[GraphExecutorState] = None):
        if state is None:
            state = GraphExecutorState()
        new_state = GraphExecutorState()
        first_node = True
        # NOTE: This logic is not yet consistent for models with multiple input nodes
        for node in self.execution_order:
            input_nodes = self.graph.find_source_nodes_of(node)
            if node.elem is None:
                continue
            out, new_state = self._apply_module(
                node,
                input_nodes,
                new_state=new_state,
                old_state=state,
                data=data if first_node else None,
            )
            new_state.cache[node.name] = out
            first_node = False

        # If the output node is a dummy nir.Output node, use the second-to-last node
        if node.name not in new_state.cache:
            node = self.execution_order[-2]
        if self.return_state:
            return new_state.cache[node.name], new_state
        else:
            return new_state.cache[node.name]



from importlib.metadata import version as metadata_version, PackageNotFoundError

try:
    __version__ = version = metadata_version("nirtorch")
    del metadata_version
except PackageNotFoundError:
    # package is not installed
    pass


import logging
from typing import Any, Callable, Dict, Optional, Sequence
import warnings

import nir
import numpy as np
import torch.nn as nn

from .graph import extract_torch_graph


DEFAULT_MAPS: Dict[nn.Module, Callable[[nn.Module], nir.NIRNode]] = {
    nn.Linear: (
        lambda module: nir.Affine(
            module.weight.detach().numpy(), module.bias.detach().numpy()
        )
    )
}


def extract_nir_graph(
    model: nn.Module,
    model_map: Callable[[nn.Module], nir.NIRNode],
    sample_data: Any,
    model_name: Optional[str] = "model",
    ignore_submodules_of=None,
    model_fwd_args=[],
    ignore_dims: Optional[Sequence[int]] = None,
) -> nir.NIRNode:
    """Given a `model`, generate an NIR representation using the specified `model_map`.

    Assumptions and known issues:
        - Cannot deal with layers like torch.nn.Identity(), since the input tensor and
          output tensor will be the same object, and therefore lead to cyclic
          connections.

    Args:
        model (nn.Module): The model of interest
        model_map (Callable[[nn.Module], nir.NIRNode]): A method that converts a given
            module type to an NIRNode type
        sample_data (Any): Sample input data to be used for model extraction
        model_name (Optional[str], optional): The name of the top level module.
            Defaults to "model".
        ignore_submodules_of (Optional[Sequence[nn.Module]]): If specified,
            the corresponding module's children will not be traversed for graph.
        ignore_dims (Optional[Sequence[int]]): Dimensions of data to be ignored for
            type/shape inference. Typically the dimensions that you will want to ignore
            are for batch and time.
    Returns:
        nir.NIR: Returns the generated NIR graph representation.
    """
    warnings.warn(
        "nirtorch.extract_nir_graph is being deprecated in favour of nirtorch.torch_to_nir. "
        "Please refer to https://neuroir.org/docs/dev_pytorch.html for detailed instructions",
        DeprecationWarning,
    )

    if len(list(model.children())):
        # If the model has submodules, ignore the top level module
        model_name = None

    # Extract a torch graph given the model
    torch_graph = extract_torch_graph(
        model, sample_data=sample_data, model_name=model_name, model_args=model_fwd_args
    )

    if ignore_submodules_of is not None:
        torch_graph = torch_graph.ignore_submodules_of(ignore_submodules_of)

    # Convert the nodes and get indices
    nir_edges = []
    input_shape = np.array(sample_data.shape)
    if ignore_dims:
        nir_nodes = {"input": nir.Input(np.delete(input_shape, ignore_dims))}
    else:
        nir_nodes = {"input": nir.Input(input_shape)}
    nir_edges = []

    subgraph_keys = []
    subgraph_input_nodekeys = []
    subgraph_output_nodekeys = []
    # Get all the NIR nodes
    for indx, node in enumerate(torch_graph.node_list):
        # Convert the node type to NIR subgraph
        mapped_node = model_map(node.elem)

        if isinstance(mapped_node, nir.NIRGraph):
            subgraph_keys.append(node.name)
            for k, v in mapped_node.nodes.items():
                # For now, we add nodes in subgraphs to the top-level node list
                # TODO: support deeper nesting -> parse graphs recursively
                assert not isinstance(v, nir.NIRGraph), "cannot handle sub-sub-graphs"

                subgraph_node_key = f"{node.name}.{k}"

                # keep track of subgraph input and outputs (to remove later)
                if isinstance(v, nir.Input):
                    subgraph_input_nodekeys.append(subgraph_node_key)
                elif isinstance(v, nir.Output):
                    subgraph_output_nodekeys.append(subgraph_node_key)

                if isinstance(v, nir.NIRNode):
                    nir_nodes[subgraph_node_key] = v
                else:
                    nir_nodes[v.name] = v  # would this ever happen??
            # Add edges from graph
            for x, y in mapped_node.edges:
                nir_edges.append((f"{node.name}.{x}", f"{node.name}.{y}"))
        else:
            nir_nodes[node.name] = mapped_node

        # Add edges from input, if first element
        # TODO: Replace with mapping to input(s)/output(s) of subgraph
        if indx == 0:  # TODO:
            keys = list(nir_nodes.keys())
            for k1, k2 in zip(keys[:-1], keys[1:]):
                nir_edges.append((k1, k2))

    # Get all the edges
    for node in torch_graph.node_list:
        for destination, shape in node.outgoing_nodes.items():
            nir_edges.append((node.name, destination.name))

        if len(node.outgoing_nodes) == 0:
            out_name = "output"
            # Try to find shape of input to the Output node
            if ignore_dims:
                out_shape = np.delete(
                    torch_graph.module_output_types[node.elem], ignore_dims
                )
            else:
                out_shape = torch_graph.module_output_types[node.elem]
            output_node = nir.Output(out_shape)
            nir_nodes[out_name] = output_node
            nir_edges.append((node.name, out_name))

    # Remove duplicate edges
    nir_edges = list(set(nir_edges))

    # change edges to subgraph to point to either input or output of subgraph
    for idx in range(len(nir_edges)):
        if nir_edges[idx][0] in subgraph_keys:
            nir_edges[idx] = (f"{nir_edges[idx][0]}.output", nir_edges[idx][1])
        if nir_edges[idx][1] in subgraph_keys:
            nir_edges[idx] = (nir_edges[idx][0], f"{nir_edges[idx][1]}.input")

    # remove subgraph input and output nodes (& redirect edges)
    for rm_nodekey in subgraph_input_nodekeys + subgraph_output_nodekeys:
        in_keys = [e[0] for e in nir_edges if e[1] == rm_nodekey]
        out_keys = [e[1] for e in nir_edges if e[0] == rm_nodekey]
        # connect all incoming to all outgoing nodes
        for in_key in in_keys:
            for out_key in out_keys:
                nir_edges.append((in_key, out_key))
        # remove the original edges
        for in_key in in_keys:
            nir_edges.remove((in_key, rm_nodekey))
        for out_key in out_keys:
            nir_edges.remove((rm_nodekey, out_key))
        # remove the node
        nir_nodes.pop(rm_nodekey)

    # HACK: remove self-connections (this is a bug in the extraction of an RNN graph)
    for edge in nir_edges:
        if edge[0] == edge[1]:
            logging.warn(f"removing self-connection {edge}")
            nir_edges.remove(edge)

    return nir.NIRGraph(nir_nodes, nir_edges)



import nir
import pytest
import torch
import torch.nn as nn
from norse.torch import LIBoxCell, LIFCell, SequentialState
from sinabs.layers import Merge

from nirtorch import extract_nir_graph
from nirtorch.graph import TorchGraph, named_modules_map, GraphTracer


class TupleModule(torch.nn.Module):
    def forward(self, data):
        return (data, data)


# Branched model
class SinabsBranchedModel(nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.relu1 = nn.ReLU()
        self.relu2_1 = nn.ReLU()
        self.relu2_2 = nn.ReLU()
        self.add_mod = Merge()
        self.relu3 = nn.ReLU()

    def forward(self, data):
        out1 = self.relu1(data)
        out2_1 = self.relu2_1(out1)
        out2_2 = self.relu2_2(out1)
        out3 = self.add_mod(out2_1, out2_2)
        out4 = self.relu3(out3)
        return out4


class NorseStatefulModel(nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.relu1 = nn.ReLU()
        self.lif = SequentialState(LIFCell())

    def forward(self, data):
        out1 = self.relu1(data)
        out2, _ = self.lif(out1)
        return out2


input_type = (2, 28, 28)
batch_size = 1

data = torch.ones((batch_size, *input_type))

my_branched_model = SinabsBranchedModel()


class DeepModel(nn.Module):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.block1 = SinabsBranchedModel()
        self.block2 = SinabsBranchedModel()

    def forward(self, data):
        out = self.block1(data)
        out2 = self.block2(out)
        return out2


mydeepmodel = DeepModel()


def test_named_modules_map():
    from nirtorch.graph import named_modules_map

    mod_map = named_modules_map(my_branched_model)
    print(mod_map)
    for k, v in mod_map.items():
        assert isinstance(k, nn.Module)
        assert isinstance(v, str)


def test_module_forward_wrapper():
    mymodel = SinabsBranchedModel()

    orig_call = nn.Module.__call__

    from nirtorch.graph import TorchGraph, module_forward_wrapper, named_modules_map

    output_types = {}
    model_graph = TorchGraph(named_modules_map(mymodel), ["block1"])
    new_call = module_forward_wrapper(model_graph, output_types)

    # Override call to the new wrapped call
    nn.Module.__call__ = new_call

    with torch.no_grad():
        _ = mymodel(data)

    # Restore normal behavior
    nn.Module.__call__ = orig_call

    print(model_graph)
    assert (
        len(model_graph.node_list) == 1 + 5 + 5 + 1
    )  # 1 top module + 5 submodules + 5 tensors + 1 output tensor
    assert len(output_types) == 6  # 1 top module + 5 submodules


def test_graph_trace_one():
    model = nn.ReLU()
    module_map = named_modules_map(model)
    sample_data = torch.rand((1,))
    with GraphTracer(module_map) as tracer, torch.no_grad():
        _ = model(sample_data)
    assert len(tracer.graph.node_list) == 1
    assert tracer.graph.node_list[0].elem == model


def test_graph_trace_many():
    from nirtorch.graph import GraphTracer, named_modules_map

    with GraphTracer(named_modules_map(my_branched_model)) as tracer, torch.no_grad():
        _ = my_branched_model(data)

    assert len(tracer.graph.node_list) == 5  # 5 submodules
    assert tracer.graph.get_edges() == [
        ("relu1", "relu2_1"),
        ("relu1", "relu2_2"),
        ("relu2_1", "add_mod"),
        ("relu2_2", "add_mod"),
        ("add_mod", "relu3"),
    ]


def test_leaf_only_graph():
    from nirtorch.graph import GraphTracer, named_modules_map

    with GraphTracer(named_modules_map(mydeepmodel)) as tracer, torch.no_grad():
        _ = mydeepmodel(data)

    print(tracer.graph)

    # Get graph with just the leaf nodes
    leaf_graph = tracer.graph.leaf_only()
    print(leaf_graph)
    assert (
        len(leaf_graph.node_list) == len(tracer.graph.node_list) - 2
    )  # No more top modules


def test_ignore_submodules_of():
    from nirtorch.graph import GraphTracer, named_modules_map

    with GraphTracer(named_modules_map(mydeepmodel)) as tracer, torch.no_grad():
        _ = mydeepmodel(data)

    top_overview_graph = tracer.graph.ignore_submodules_of(
        [SinabsBranchedModel]
    ).leaf_only()
    assert len(top_overview_graph.node_list) == 2


def test_snn_branched():
    from sinabs.layers import ChannelShift, IAFSqueeze, SumPool2d
    from torch.nn import Conv2d

    from nirtorch.graph import extract_torch_graph

    class MySNN(nn.Module):
        def __init__(self) -> None:
            super().__init__()
            self.conv1 = Conv2d(2, 8, 3, bias=False)
            self.iaf1 = IAFSqueeze(batch_size=1)
            self.pool1 = SumPool2d(2)
            self.conv2_1 = Conv2d(8, 16, 3, stride=1, padding=1, bias=False)
            self.iaf2_1 = IAFSqueeze(batch_size=1)
            self.pool2_1 = SumPool2d(2)
            self.conv2_2 = Conv2d(8, 16, 5, stride=1, padding=2, bias=False)
            self.iaf2_2 = IAFSqueeze(batch_size=1)
            self.pool2_2 = SumPool2d(2)
            self.channel_shift = ChannelShift(channel_shift=16)
            self.merge = Merge()
            self.conv3 = Conv2d(32, 10, 3, stride=3, bias=False)
            self.iaf3 = IAFSqueeze(batch_size=1)

        def forward(self, spikes):
            out = self.conv1(spikes)
            out = self.iaf1(out)
            out = self.pool1(out)

            out1 = self.conv2_1(out)
            out1 = self.iaf2_1(out1)
            out1 = self.pool2_1(out1)

            out2 = self.conv2_2(out)
            out2 = self.iaf2_2(out2)
            out2 = self.pool2_2(out2)

            out = self.merge(out1, self.channel_shift(out2))
            out = self.conv3(out)
            out = self.iaf3(out)
            return out

    my_snn = MySNN()
    graph = extract_torch_graph(
        my_snn, sample_data=torch.rand((100, 2, 14, 14)), model_name=None
    )

    assert len(graph.node_list) == 13


def test_input_output():
    from norse.torch import to_nir as norse_to_nir

    g = norse_to_nir(NorseStatefulModel(), data)
    assert len(g.nodes) == 4  # in -> relu -> lif -> out
    assert len(g.edges) == 3


def test_output_type_when_single_node():
    g = extract_nir_graph(
        torch.nn.ReLU(),
        lambda x: nir.Threshold(torch.tensor(0.1)),
        sample_data=torch.rand((1,)),
    )
    assert "input" in g.nodes
    assert "output" in g.nodes
    g.nodes["output"].output_type["output"] == torch.Size([1])


def test_sequential_flatten():
    d = torch.empty(3, 4)
    g = extract_nir_graph(torch.nn.Flatten(1), lambda x: nir.Flatten(d.shape, 1), d)
    assert tuple(g.nodes["input"].input_type["input"]) == (3, 4)

    d = torch.empty(2, 3, 4)
    g = extract_nir_graph(
        torch.nn.Flatten(1), lambda x: nir.Flatten(d.shape, 1), d, ignore_dims=[0]
    )
    assert tuple(g.nodes["input"].input_type["input"]) == (3, 4)


@pytest.mark.skip(reason="Not supported yet")
def test_captures_recurrence_automatically():
    class Recmodel(torch.nn.Module):
        def __init__(self):
            super().__init__()
            self.r = LIBoxCell()
            self.l = torch.nn.Linear(1, 1)

        def forward(self, x, state=None):
            if state is None:
                state = [None, torch.zeros_like(x)]
            s1, s2 = state  # s1 = LIF, s2 = recurrence
            out, s1 = self.r(x + s2, s1)
            s2 = self.l(out)
            return out, [s1, s2]

    from norse.torch.utils.export_nir import _extract_norse_module

    m = Recmodel()
    data = torch.randn((1, 1))
    actual, actual_state = m(*m(data))
    d = extract_nir_graph(m, _extract_norse_module, data)
    assert d.nodes.keys() == {"input", "l", "r", "output"}
    assert set(d.edges) == {("input", "r"), ("r", "l"), ("l", "output"), ("r", "r")}


@pytest.mark.skip(reason="Subgraphs are currently flattened")
def test_captures_recurrence_manually():
    def export_affine_rec_gru(module):
        if isinstance(module, torch.nn.Linear):
            return nir.Affine(module.weight, module.bias)
        elif isinstance(module, torch.nn.ReLU):
            return nir.NIRGraph(nodes={"i": nir.I(torch.randn(1))}, edges=[("i", "i")])
        else:
            raise ValueError(f"Unsupported module {module}")

    m = torch.nn.Sequential(torch.nn.Linear(1, 1), torch.nn.ReLU())
    data = torch.randn((1, 1))

    d = extract_nir_graph(m, export_affine_rec_gru, data)
    assert d.nodes.keys() == {"input", "0", "1.i", "output"}
    assert set(d.edges) == {("input", "0"), ("0", "1"), ("1.i", "1.i"), ("1", "output")}
    # TODO: Pass recursively
    # assert d.nodes["1"].nodes.keys() == {"i"}
    # assert d.nodes["1"].edges == [("1.i", "1.i")]


def test_graph_captures_input():
    model = nn.Sequential(
        nn.Conv2d(2, 8, 3),
        nn.AvgPool2d(2),
        nn.ReLU(),
    )

    modules = {module: name for name, module in list(model.named_modules())[1:]}
    g = TorchGraph(modules, ["0"])
    assert len(g.inputs) == 1
    assert g.inputs[0] == "0"



import nir
import numpy as np
import torch

import nirtorch

use_snntorch = False
# use_snntorch = True


if use_snntorch:
    import snntorch as snn


def _nir_to_snntorch_module(node: nir.NIRNode) -> torch.nn.Module:
    if isinstance(node, (nir.Linear, nir.Affine)):
        return torch.nn.Linear(*node.weight.shape)

    elif isinstance(node, (nir.LIF, nir.CubaLIF)):
        return snn.Leaky(0.9, init_hidden=True)

    else:
        return None


def _nir_to_pytorch_module(node: nir.NIRNode) -> torch.nn.Module:
    if isinstance(node, (nir.Linear, nir.Affine)):
        return torch.nn.Linear(*node.weight.shape)

    elif isinstance(node, (nir.LIF, nir.CubaLIF)):
        return torch.nn.Linear(1, 1)

    else:
        return None


if use_snntorch:
    _nir_to_torch_module = _nir_to_snntorch_module
else:
    _nir_to_torch_module = _nir_to_pytorch_module


def _create_torch_model() -> torch.nn.Module:
    if use_snntorch:
        return torch.nn.Sequential(
            torch.nn.Linear(1, 1), snn.Leaky(0.9, init_hidden=True)
        )
    else:
        return torch.nn.Sequential(torch.nn.Linear(1, 1), torch.nn.Identity())


def _torch_to_nir(module: torch.nn.Module) -> nir.NIRNode:
    if isinstance(module, torch.nn.Linear):
        return nir.Linear(np.array(module.weight.data))

    else:
        return None


def _lif_nir_graph(from_file=True):
    if from_file:
        return nir.read("tests/lif_norse.nir")
    else:
        return nir.NIRGraph(
            nodes={
                "input": nir.Input(input_type={"input": np.array([1])}),
                "0": nir.Affine(weight=np.array([[1.0]]), bias=np.array([0.0])),
                "1": nir.LIF(
                    tau=np.array([0.1]),
                    r=np.array([1.0]),
                    v_leak=np.array([0.0]),
                    v_threshold=np.array([0.1]),
                ),
                "output": nir.Output(output_type={"output": np.array([1])}),
            },
            edges=[("input", "0"), ("0", "1"), ("1", "output")],
        )


def test_nir_to_torch_to_nir(from_file=True):
    graph = _lif_nir_graph(from_file=from_file)
    assert graph is not None
    module = nirtorch.load(graph, _nir_to_torch_module)
    assert module is not None
    graph2 = nirtorch.extract_nir_graph(module, _torch_to_nir, torch.zeros(1, 1))
    edges1 = sorted(graph.edges)
    edges2 = sorted(graph2.edges)
    for e1, e2 in zip(edges1, edges2):
        assert e1 == e2


# if __name__ == '__main__':
#     test_nir_to_torch_to_nir(from_file=False)



import collections
import typing

import nir
import torch
import inspect

NodeMapType = typing.Dict[nir.NIRNode, typing.Callable[[nir.NIRNode], torch.nn.Module]]


def _default_map_affine(linear: nir.Affine) -> torch.nn.Linear:
    module = torch.nn.Linear(
        linear.weight.shape[-1], linear.weight.shape[-2], bias=True
    )
    module.weight.data = torch.from_numpy(linear.weight)
    module.weight.bias = torch.from_numpy(linear.bias)
    return module


def _default_map_conv1d(conv: nir.Conv1d) -> torch.nn.Conv1d:
    module = torch.nn.Conv1d(
        conv.weight.shape[-3],
        conv.weight.shape[-2],
        conv.weight.shape[-1],
        stride=conv.stride,
        padding=conv.padding,
        dilation=conv.dilation,
        groups=conv.groups,
        bias=True,
    )
    module.weight.data = torch.from_numpy(conv.weight)
    module.bias.data = torch.from_numpy(conv.bias)
    return module


def _default_map_conv2d(conv: nir.Conv2d) -> torch.nn.Conv2d:
    c = torch.nn.Conv2d(
        conv.weight.shape[-4],
        conv.weight.shape[-3],
        conv.weight.shape[-2:],
        stride=conv.stride,
        padding=conv.padding,
        dilation=conv.dilation,
        groups=conv.groups,
        bias=True,
    )
    c.weight.data = torch.from_numpy(conv.weight)
    c.bias.data = torch.from_numpy(conv.bias)
    return c


def _default_map_linear(linear: nir.Linear) -> torch.nn.Linear:
    module = torch.nn.Linear(
        linear.weight.shape[-1], linear.weight.shape[-2], bias=False
    )
    module.weight.data = torch.from_numpy(linear.weight)
    return module


DEFAULT_MAP: NodeMapType = {
    nir.Input: lambda input: torch.nn.Identity(),
    nir.Output: lambda output: torch.nn.Identity(),
    # Node mappings
    nir.Affine: _default_map_affine,
    nir.Conv1d: _default_map_conv1d,
    nir.Conv2d: _default_map_conv2d,
    nir.Linear: _default_map_linear,
}


def _map_nir_node_to_torch(
    node: nir.NIRNode, node_map: NodeMapType
) -> typing.Optional[torch.nn.Module]:
    if type(node) in node_map:
        return node_map[type(node)](node)
    else:
        raise ValueError(
            f"Unknown node type {type(node)}, mapping does not exist in node_map"
        )


def _construct_module_dict_recursive(
    nir_graph: nir.NIRGraph, node_map: NodeMapType
) -> torch.nn.ModuleDict:
    owning_module = torch.nn.ModuleDict()
    for name, node in nir_graph.nodes.items():
        # Recurse into subgraphs
        if isinstance(node, nir.NIRGraph):
            owning_module[name] = _construct_module_dict_recursive(node, node_map)
        else:
            mapped_module = _map_nir_node_to_torch(node, node_map=node_map)
            if mapped_module is not None:
                owning_module[name] = mapped_module
    return owning_module


def _find_input_nodes(
    name: str,
    edges: typing.List[typing.Tuple[str, str]],
    node_outputs: typing.Dict[str, torch.fx.Node],
):
    """
    Looks through the edges and find nodes that are connected to the given node as inputs.
    If one of the inputs doesn't exist because it hasn't been defined, we return None.

    # TODO: Multiple inputs should be added together
    """
    input_node_names = list(filter(lambda t: t[1] == name, edges))
    output_nodes = tuple(node_outputs.get(a, None) for a, _ in input_node_names)
    return output_nodes


def _is_stateful(module: torch.nn.Module) -> bool:
    signature = inspect.signature(module.forward)
    return (
        "state" in signature.parameters  # Contains a state input argument
        or isinstance(module, torch.nn.RNNBase)  # RNN modules are stateful
        or isinstance(
            module, torch.fx.GraphModule
        )  # GraphModules created by us are always stateful
    )


def _construct_state_recursive(
    parent_module: torch.nn.ModuleDict,
) -> typing.Dict[str, typing.Any]:
    state = {}
    for name, module in parent_module.items():
        if isinstance(module, torch.nn.ModuleDict):
            state[name] = _construct_state_recursive(module)
        else:
            state[name] = None
    return state


def _construct_fx_graph(
    owning_module: torch.nn.ModuleDict, nir_graph: nir.NIRGraph
) -> torch.fx.GraphModule:
    node_outputs = {}
    recursion_counter = collections.Counter(nir_graph.nodes.keys())
    # The maximum iterations per node (see https://github.com/neuromorphs/NIRTorch/pull/28#discussion_r1959343951)
    max_iterations = min(3, len(recursion_counter))
    torch_graph = torch.fx.Graph(owning_module)
    # Create a queue of the modules where we can re-insert modules for re-processing
    # in case of self-reference or if the graph is given out of order
    module_queue = collections.deque(nir_graph.nodes.items())

    # Create some dummy input for use when processing recursive elements
    dummy_input = torch_graph.call_function(torch.ones, (1,))

    # Loop through all the nodes in the queue
    while module_queue:
        module_name, module = module_queue.popleft()
        if recursion_counter[module_name] > max_iterations:
            raise RecursionError(
                f"Module {module_name} has been traversed multiple times"
                " which may be a bug in the graph or in the implementation."
                " Please file an issue at github.com/neuromorphs/nirtorch"
            )

        if isinstance(module, nir.Input):
            if len(module.output_type) > 1:
                raise NotImplementedError("Multiple inputs are currently not supported")
            for input_name, _ in module.input_type.items():
                node_outputs[module_name] = torch_graph.create_node(
                    "placeholder", input_name
                )

            # Create a dummy state variable for each module
            default_state = _construct_state_recursive(owning_module)
            # - This has to happen *after* we create the input placeholders to avoid adding a parameter with default values before the input argument
            state_argument = torch_graph.placeholder(
                "state",
                type_expr=typing.Dict[str, typing.Any],
                default_value=default_state,
            )
        elif isinstance(module, nir.Output):
            # Only add the output if the module is the final node in the queue
            # - Output nodes are essentially return statements
            if len(module_queue) > 0:
                module_queue.append((module_name, module))
            else:
                if len(module.output_type) > 1:
                    raise ValueError("Multiple outputs are currently not supported")
                for input_name, output in module.input_type.items():
                    # First fetch the required input nodes
                    module_input_nodes = _find_input_nodes(
                        module_name, edges=nir_graph.edges, node_outputs=node_outputs
                    )
                    # If the module uses input that is not yet defined, set the inputs to some dummy value
                    # and enqueue the module again for processing (where it's hopefully defined)
                    if None in module_input_nodes:
                        module_input_nodes = [dummy_input for _ in module_input_nodes]
                        module_queue.append((module_name, module))

                    # # If we visited the node before, simply update the new inputs
                    # if recursion_counter[module_name] > 1:
                    #     output_node = node_outputs[module_name]
                    #     for index, input_node in enumerate(module_input_nodes):
                    #         output_node.update_arg(index, input_node)
                    # If it has not been visited before, add the node
                    node_outputs[module_name] = torch_graph.output(
                        (*module_input_nodes, state_argument)
                    )
        else:
            # 1. Recursively wire subgraphs
            if isinstance(module, nir.NIRGraph):
                owning_module[module_name] = _construct_fx_graph(
                    owning_module[module_name], module
                )

            # 2. Call the module
            kwargs = {}
            # - If the module has a state parameter, provide a state argument and store the state
            is_stateful = _is_stateful(owning_module[module_name])
            if is_stateful:
                kwargs["state"] = torch_graph.call_method(
                    "__getitem__", (state_argument, module_name)
                )
            # - If the module uses input that is not yet defined, set the input to some dummy values
            #   and enqueue the module again for processing (where it's hopefully defined)
            module_input_nodes = _find_input_nodes(
                module_name, nir_graph.edges, node_outputs=node_outputs
            )
            if None in module_input_nodes:
                # This module depends on another module that hasn't yet been constructed
                module_input_nodes = tuple([dummy_input for _ in module_input_nodes])
                # Enqueue for later processing
                module_queue.append((module_name, module))
            # Otherwise, call the module as usual
            else:
                output = torch_graph.call_module(
                    module_name,
                    module_input_nodes,
                    kwargs,
                )
                # - If the module is stateful, we assume the second part is a state object and must separate that from the output
                if is_stateful:
                    # Add the raw output to the graph
                    node_outputs[f"{module_name}_raw"] = output
                    # Add the module state to the graph for use as the input to the next module
                    node_outputs[f"{module_name}"] = torch_graph.call_method(
                        "__getitem__", (output, 0)
                    )
                    # Add the state to the state dictionary
                    state_node = node_outputs[f"{module_name}_state"] = (
                        torch_graph.call_method("__getitem__", (output, 1))
                    )
                    torch_graph.call_method(
                        "__setitem__", (state_argument, module_name, state_node)
                    )
                # - If the output is not a tuple, return as normal
                else:
                    node_outputs[module_name] = output

        recursion_counter[module_name] += 1

    # Ensure correctness
    torch_graph.lint()

    return torch.fx.GraphModule(owning_module, torch_graph)


def nir_to_torch(
    nir_graph: nir.NIRGraph,
    node_map: NodeMapType,
    default_map: NodeMapType = DEFAULT_MAP,
) -> torch.fx.GraphModule:
    """
    Maps a NIRGraph as an executable PyTorch GraphModule (torch.fx.GraphModule).
    We first map all individual nodes using the node_map, where a common set of mappings are provided by default (e. g. Linear, Conv, etc.)
    Then, we wire all the nodes together into an executable torch.fx.GraphModule.
    Finally, we wrap the execution in a StatefulInterpreter, to ensure that the internal state of modules are handled correctly.
    """
    map_with_defaults = dict(default_map)
    map_with_defaults.update(node_map)  # Overwrite defaults with node_map
    # First convert all nodes into a module dictionary
    owning_module = _construct_module_dict_recursive(nir_graph, map_with_defaults)
    # Then wire the graph recursively
    return _construct_fx_graph(owning_module=owning_module, nir_graph=nir_graph)



import warnings
from numbers import Number
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Type, Union

import nir
import torch
import torch.nn as nn

from .utils import sanitize_name


def named_modules_map(
    model: nn.Module, model_name: Optional[str] = "model"
) -> Dict[str, nn.Module]:
    """Inverse of named modules dictionary.

    Args:
        model (nn.Module): The module to be hashed
        model_name (str | None): Name of the top level module. If this doesn't need
            to be include, this option can be set to None

    Returns:
        Dict[str, nn.Module]: A dictionary with modules as keys, and names as values
    """
    modules_map = {}
    for name, mod in model.named_modules():
        # Ignore sequential modules
        if isinstance(mod, nn.Sequential):
            continue
        modules_map[mod] = name
    if model_name is None:
        if model in modules_map:
            del modules_map[model]
    else:
        modules_map[model] = model_name
    return modules_map


class Node:
    def __init__(
        self,
        elem: Any,
        name: str,
        outgoing_nodes: Optional[Dict["Node", torch.Tensor]] = None,
    ) -> None:
        self.elem = elem
        self.name = name
        if not outgoing_nodes:
            self.outgoing_nodes = {}
        else:
            self.outgoing_nodes = outgoing_nodes

    def add_outgoing(self, node: "Node", shape=None) -> None:
        self.outgoing_nodes[node] = shape

    def __str__(self) -> str:
        return f"Node: {self.name} ({type(self.elem)}), Out: {len(self.outgoing_nodes)}"

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, other: Any) -> bool:
        # Two nodes are meant to be the same if they refer to the same element
        try:
            return self.elem is other.elem
        except AttributeError:
            return False

    def __hash__(self):
        # Two nodes are same if they reference the same element
        return hash(self.elem)


def _find_input_nodes(graph: "TorchGraph") -> Set[nir.Input]:
    # First, check if the graph is empty or singular
    if len(graph.node_list) == 0:
        raise ValueError("Cannot find the input of an empty graph")
    elif len(graph.node_list) == 1:
        return set(graph.node_list)

    # Then, retrace the graph to find "orphan" nodes
    candidates = set(graph.node_list)
    for node in graph.node_list:
        outgoing = set(node.outgoing_nodes.keys())
        candidates = candidates - outgoing
    return candidates


class TorchGraph:
    def __init__(
        self,
        module_names: Dict[nn.Module, str],
        inputs: Optional[Set[str]],
        module_output_types: Dict[nn.Module, torch.Tensor] = {},
    ) -> None:
        self.module_names = module_names
        self.node_list: List[Node] = []
        self.module_output_types = module_output_types
        self._last_used_tensor_id = None
        self.inputs = inputs

    def _discover_inputs(self, edges: List[Tuple[str, str]]):
        nodes = {name: node for node, name in self.module_names.items()}
        for src, dst in edges:
            # Allow edges to refer to subgraph inputs and outputs
            if src not in nodes and f"{src}.output" in nodes:
                src = f"{src}.output"
            if dst not in nodes and f"{dst}.input" in nodes:
                dst = f"{dst}.input"

            self.add_edge(nodes[src], nodes[dst])
        self.inputs = _find_input_nodes(self)

    @staticmethod
    def from_torch_modules(
        nodes: Dict[str, nn.Module], edges: List[Tuple[str, str]]
    ) -> "TorchGraph":
        module_names = {module: sanitize_name(name) for name, module in nodes.items()}
        # Construct the graph by adding edges and, finally, inputs
        graph = TorchGraph(module_names=module_names, inputs=set())
        graph._discover_inputs(edges)
        return graph

    def num_edges(self) -> int:
        count = 0
        for node in self.node_list:
            count += node.outgoing_nodes
        return count

    def get_unique_tensor_id(self) -> str:
        if self._last_used_tensor_id is None:
            self._last_used_tensor_id = 0
        else:
            self._last_used_tensor_id += 1
        return str(self._last_used_tensor_id)

    def __contains__(self, elem: Union[torch.Tensor, nn.Module]) -> bool:
        for node in self.node_list:
            if elem is node.elem:
                return True
        return False

    def add_elem(self, elem, name: str) -> Node:
        if elem in self:
            warnings.warn(f"{name}: Node already exists for this element ")
            return self.find_node(elem)
        else:
            node = Node(elem, name)
            self.node_list.append(node)
            return node

    def add_or_get_node_for_elem(self, elem: Union[torch.Tensor, nn.Module]):
        if elem in self:
            return self.find_node(elem)
        else:
            # Generate a name
            if elem in self.module_names:
                name = self.module_names[elem]
            else:
                if isinstance(elem, Number):
                    elem = torch.as_tensor(elem)
                if not isinstance(elem, torch.Tensor):
                    raise ValueError(f"Unknown element type {type(elem)}")
                name = f"Tensor_{self.get_unique_tensor_id()}{tuple(elem.shape)}"
            # add and return the node
            new_node = self.add_elem(elem, name)
            return new_node

    def find_node(self, elem: Union[torch.Tensor, nn.Module]):
        for node in self.node_list:
            if elem is node.elem:
                return node
        raise ValueError("elem not found")

    def add_edge(
        self,
        source: Union[torch.Tensor, nn.Module],
        destination: Union[torch.Tensor, nn.Module],
        shape: torch.Tensor = None,
    ):
        if self._is_mod_and_not_in_module_names(source):
            return
        if self._is_mod_and_not_in_module_names(destination):
            return

        if source is None or destination is None:
            return  # Stateful models may have Nones

        source_node = self.add_or_get_node_for_elem(source)
        destination_node = self.add_or_get_node_for_elem(destination)
        source_node.add_outgoing(destination_node, shape)
        return source_node, destination_node

    def get_leaf_modules(self) -> Dict[nn.Module, str]:
        filtered_module_names = {}

        for mod, _ in self.module_names.items():
            # Add module to dict
            filtered_module_names[mod] = self.module_names[mod]
            child_in_graph = False
            for _, submod in mod.named_children():
                if submod in self:
                    child_in_graph = True
                    break
            if child_in_graph:
                del filtered_module_names[mod]
        return filtered_module_names

    def _is_mod_and_not_in_module_names(self, elem: Any) -> bool:
        """Check if a node is a module and is included in the module_names of this
        graph.

        Args:
            node (Node): Node to verify

        Returns:
            bool
        """
        if isinstance(elem, nn.Module) and elem not in self.module_names:
            return True
        else:
            return False

    def populate_from(self, other_graph: "TorchGraph"):
        self.module_output_types.update(other_graph.module_output_types)
        for node in other_graph.node_list:
            for outgoing_node, shape in node.outgoing_nodes.items():
                self.add_edge(node.elem, outgoing_node.elem, shape)

    def __str__(self) -> str:
        return self.to_md()

    def debug_str(self) -> str:
        debug_str = ""
        for node in self.node_list:
            debug_str += f"{node.name} ({node.elem.__class__.__name__})\n"
            for outgoing, shape in node.outgoing_nodes.items():
                debug_str += (
                    f"\t-> {outgoing.name} ({outgoing.elem.__class__.__name__})\n"
                )
        return debug_str.strip()

    def to_md(self) -> str:
        mermaid_md = """```mermaid\ngraph TD;\n"""
        for node in self.node_list:
            if node.outgoing_nodes:
                for outgoing, _ in node.outgoing_nodes.items():
                    mermaid_md += f"{node.name} --> {outgoing.name};\n"
            else:
                mermaid_md += f"{node.name};\n"

        return mermaid_md + "\n```\n"

    def leaf_only(self) -> "TorchGraph":
        leaf_modules = self.get_leaf_modules()
        filtered_graph = TorchGraph(leaf_modules, inputs=self.inputs)
        # Populate edges
        filtered_graph.populate_from(self)
        return filtered_graph

    def ignore_submodules_of(self, classes: List[Type]) -> "TorchGraph":
        new_named_modules = {}

        # Gather a list of all top level modules, whose submodules are to be ignored
        top_level_modules: List[nn.Module] = []
        for mod in self.module_names.keys():
            if mod.__class__ in classes:
                top_level_modules.append(mod)

        # List all the submodules of the above module list
        sub_modules_to_ignore: List[nn.Module] = []
        for top_mod in top_level_modules:
            for sub_mod in top_mod.modules():
                if sub_mod is not top_mod:
                    sub_modules_to_ignore.append(sub_mod)

        # Iterate over all modules and check if they are submodules of the above list
        for mod, name in self.module_names.items():
            if mod not in sub_modules_to_ignore:
                new_named_modules[mod] = name
        # Create a new graph with the allowed modules
        new_graph = TorchGraph(
            new_named_modules,
            inputs=self.inputs,
            module_output_types=self.module_output_types,
        )
        new_graph.populate_from(self)
        return new_graph

    def find_source_nodes_of(self, node: Node) -> List[Node]:
        """Find all the sources of a node in the graph.

        Args:
            node (Node): Node of interest

        Returns:
            List[Node]: A list of all nodes that have this node as outgoing_node
        """
        source_node_list = []
        for source_node in self.node_list:
            for outnode, _ in source_node.outgoing_nodes.items():
                if node == outnode:
                    source_node_list.append(source_node)
        return source_node_list

    def ignore_tensors(self) -> "TorchGraph":
        """Simplify the graph by ignoring all the tensors in it.

        Returns:
            Graph: Returns a simplified graph with only modules in it
        """
        return self.ignore_nodes(torch.Tensor)

    def ignore_nodes(self, class_type: Type) -> "TorchGraph":
        # Filter module names to remove the given class type
        new_module_names = {
            k: v for k, v in self.module_names.items() if not isinstance(k, class_type)
        }

        # Generate the new graph with the filtered module names
        graph = TorchGraph(
            new_module_names,
            inputs=self.inputs,
            module_output_types=self.module_output_types,
        )
        # Iterate over all the nodes
        for node in self.node_list:
            if isinstance(node.elem, class_type):
                # Get its source
                source_node_list = self.find_source_nodes_of(node)
                if len(source_node_list) == 0:
                    # If no source, this is probably origin node, just drop it
                    continue
                # Get all of its destinations
                if node.outgoing_nodes:
                    # If no destinations, it is a leaf node, just drop it.
                    for outgoing_node, shape in node.outgoing_nodes.items():
                        # Directly add an edge from source to destination
                        for source_node in source_node_list:
                            graph.add_edge(source_node.elem, outgoing_node.elem, shape)
                            # NOTE: Assuming that the destination is not of the same
                            # type here
            else:
                # This is to preserve the graph if executed on a graph that is
                # already filtered
                for outnode, shape in node.outgoing_nodes.items():
                    if not isinstance(outnode.elem, class_type):
                        graph.add_edge(node.elem, outnode.elem, shape)
        return graph

    def get_edges(self) -> List[Tuple[str, str]]:
        edges = []
        for node in self.node_list:
            for outgoing_node in node.outgoing_nodes:
                edges.append((node.name, outgoing_node.name))
        return edges

    def get_root(self) -> List[Node]:
        """Returns the root node/s of the graph.

        Returns:
            List[Node]: A list of root nodes for the graph.
        """
        return self.inputs


_torch_module_call = torch.nn.Module.__call__


def module_forward_wrapper(
    model_graph: TorchGraph, output_types: Dict[nn.Module, torch.Tensor]
) -> Callable[..., Any]:
    def my_forward(mod: nn.Module, *args, **kwargs) -> Any:
        out = _torch_module_call(mod, *args, **kwargs)

        if isinstance(out, tuple):
            out_tuple = (out[0],)
            output_types[mod] = out[0].shape
        elif isinstance(out, torch.Tensor):
            out_tuple = (out,)
            output_types[mod] = out.shape
        else:
            raise Exception("Unknown output format")

        # Iterate over all inputs
        for i, input_data in enumerate(args):
            # Create nodes and edges
            model_graph.add_edge(
                input_data,
                mod,
                input_data.shape if isinstance(input_data, torch.Tensor) else None,
            )

        # Iterate over all outputs and create nodes and edges
        for output_data in out_tuple:
            # Create nodes and edges
            model_graph.add_edge(
                mod,
                output_data,
                output_data.shape if isinstance(output_data, torch.Tensor) else None,
            )
        return out

    return my_forward


class GraphTracer:
    """Context manager to trace a model's execution graph.

    Example:

    ```python
    with GraphTracer(mymodel) as tracer, torch.no_grad():
        out = mymodel(data)

    print(tracer.graph.to_md())
    ```
    """

    def __init__(self, module_map: Dict[nn.Module, str]) -> None:
        self.original_torch_call = nn.Module.__call__
        self.output_types = {}
        self.graph = TorchGraph(module_map, self.output_types)

    def __enter__(self) -> "GraphTracer":
        # Override the torch call method
        nn.Module.__call__ = module_forward_wrapper(self.graph, self.output_types)
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        # Restore normal behavior
        nn.Module.__call__ = self.original_torch_call
        # Prune tensor edges
        pruned_graph = self.graph.ignore_tensors()
        # Cover the edgse where a graph has a single node
        # by manually re-creating the graph
        if (
            len(pruned_graph.node_list) == 0
            and len(self.graph.node_list) == 3  # (2 tensors + 1 module)
        ):
            module, name = next(iter(self.graph.module_names.items()))
            self.graph = TorchGraph(self.graph.module_names, [name])
            self.graph.node_list.append(Node(module, name))
        else:
            self.graph = pruned_graph
        # Rediscover inputs to ensure correct connectivity
        self.graph._discover_inputs(self.graph.get_edges())


def extract_torch_graph(
    model: nn.Module,
    sample_data: Any,
    model_name: Optional[str] = "model",
    model_args=[],
) -> TorchGraph:
    """Extract computational graph between various modules in the model
    NOTE: This method is not capable of any compute happening outside of module
    definitions.

    Args:
        model (nn.Module): The module to be analysed
        sample_data (Any): Sample data to be used to run by the model
        model_name (Optional[str], optional): Name of the top level module.
          If specified, it will be included in the graph.
          If set to None, only its submodules will be listed in the graph.
          Defaults to "model".
            for n in torch_graph.node_list:
                n_names = {x.name for x in n.outgoing_nodes}
                if node.name in n_names:
                    shape = n.outgoing_nodes[node]
    Returns:
        Graph: A graph object representing the computational graph of the given model
    """
    module_map = named_modules_map(model, model_name=model_name)
    with GraphTracer(module_map) as tracer, torch.no_grad():
        _ = model(sample_data, *model_args)

    # HACK: The current graph is using copy-constructors, that detaches
    # the traced output_types from the original graph.
    # In the future, find a way to synchronize the two representations
    tracer.graph.module_output_types = tracer.output_types

    return tracer.graph



import torch

from nirtorch.graph import TorchGraph
from nirtorch.graph_executor import GraphExecutor


def test_execute_stateful_then_stateless():
    class StatefulModel(torch.nn.Module):
        def __init__(self):
            super().__init__()

        def forward(self, x, state=None):
            if state is None:
                state = 0
            return x + state, x

    s = StatefulModel()
    lin = torch.nn.Linear(1, 1, bias=False)
    lin.weight.data = torch.tensor([[1]]).float()
    graph = TorchGraph.from_torch_modules({"0": s, "1": lin}, [("0", "1")])

    # With state
    executor = GraphExecutor(graph, True)
    data = torch.ones(1, 1)
    out, state = executor(data)
    assert torch.allclose(out, torch.ones(1))
    out, _ = executor(data, state)
    assert torch.allclose(out, torch.ones(1) * 2)

    # Without state
    executor = GraphExecutor(graph, False)
    data = torch.ones(1, 1)
    out = executor(data)
    assert torch.allclose(out, torch.ones(1))
